from typing import Annotated, Iterable, Optional

from colorama import Fore, Style

from .abstract import LogDownloader
from .constants import DEFAULT_FILES, DEFAULT_NUMBER_OF_ROUNDS, DEFAULT_ONLY_PLAYED, DEFAULT_CKEY_OUTPUT_PATH
from ..byond import canonicalize
from .. import scrubby
from ..constants import NEGATIVE_RESPONSES, POSITIVE_RESPONSES


class CkeyLogDownloaderException(Exception):
    """Uh oh, something went wrong"""


class CkeyLogDownloader(LogDownloader):
    """Downloads logs in which a ckey was present"""
    ckey: Annotated[Optional[str], "Canonical form of user's key, can be None"]
    only_played: Annotated[bool, "If ckey is set dictates if the log downloader only counts player rounds"]\
        = DEFAULT_ONLY_PLAYED
    number_of_rounds: Annotated[int, "The number of rounds to download"] = DEFAULT_NUMBER_OF_ROUNDS
    output_path: Annotated[str, "Where should we write the file to?"] = DEFAULT_CKEY_OUTPUT_PATH.format(ckey="output")
    filter_logs: Annotated[bool, "Should we filter the logs?"] = True

    def __init__(self, key: str = None, only_played: bool = DEFAULT_ONLY_PLAYED,
                 number_of_rounds: int = DEFAULT_NUMBER_OF_ROUNDS, output_path: str = None) -> None:
        self.ckey = canonicalize(key) if key else None
        self.only_played = only_played
        self.number_of_rounds = number_of_rounds
        self.output_path = output_path.format(ckey=self.ckey or "output")

    async def update_round_list(self) -> None:
        self.rounds = await scrubby.GetReceipts(self.ckey, self.number_of_rounds, self.only_played)

    def filter_lines(self, logs: Iterable[bytes]) -> Iterable[bytes]:
        if not self.filter_logs:
            for log in logs:
                yield log
        if not self.ckey:
            raise CkeyLogDownloaderException("Ckey was empty")
        ckey = self.ckey.lower().encode('utf-8')
        for log in logs:
            if ckey in log.lower():
                yield log

    @staticmethod
    def interactive() -> LogDownloader:
        ckey = input("CKEY: ").strip()
        while True:
            number_of_rounds = input(f"How many rounds? [{DEFAULT_NUMBER_OF_ROUNDS}] ")
            try:
                if not number_of_rounds.strip():
                    number_of_rounds = DEFAULT_NUMBER_OF_ROUNDS
                    break
                number_of_rounds = int(number_of_rounds) if number_of_rounds.isdigit() else int(number_of_rounds, 16)
                break
            except ValueError:
                print(f"{Fore.RED}Rounds should be an int in base 10 or 16{Fore.RESET}")
        print(f"Do you want to get only rounds in which they played? [y/{Style.BRIGHT}N{Style.RESET_ALL}] ", end="")
        only_played = input().lower() in POSITIVE_RESPONSES  # Input and colorama don't mix
        output_path = input(f"Where should I write the file? [{DEFAULT_CKEY_OUTPUT_PATH}] ")
        output_path = output_path or DEFAULT_CKEY_OUTPUT_PATH.format(ckey=ckey)
        downloader = CkeyLogDownloader(key=ckey, only_played=only_played,
                                       number_of_rounds=number_of_rounds, output_path=output_path)
        print("Which files do you want to download?")
        print("(separate the files with a comma, like so: attack.txt,game.txt,pda.txt)")
        file_list = [x.strip() for x in input(f"[{','.join(DEFAULT_FILES)}] ").split(',') if x.strip()]
        if file_list:
            downloader.files = file_list
        print(f"Want only the logs that this person is in? [{Style.BRIGHT}Y{Style.RESET_ALL}/n] ", end="")
        downloader.filter_logs = input().lower().strip() not in NEGATIVE_RESPONSES
        downloader.try_authenticate_interactive()
        return downloader



GAME_TXT_URL = "https://tgstation13.org/parsed-logs/{server}/data/logs/{year}/{month}/{day}/round-{round_id}/{file_name}"
GAME_TXT_ADMIN_URL = "https://tgstation13.org/raw-logs/{server}/data/logs/{year}/{month}/{day}/round-{round_id}/{file_name}"
DEFAULT_FILES = ["game.txt"]
DEFAULT_NUMBER_OF_ROUNDS = 150
DEFAULT_OUTPUT_PATH = "logs.txt"
DEFAULT_CKEY_OUTPUT_PATH = "{ckey}.txt"
DEFAULT_ROUND_OUTPUT_PATH = "{start}-{end}.txt"
DEFAULT_ONLY_PLAYED = False



"""
Represents an abstract item on the menu.
Do not import this class directly, import it from .menu instead.
"""
from abc import ABC, abstractmethod
from typing import Annotated


class MenuItem(ABC):
    """Represents an item on the main menu"""

    weight: Annotated[float, "The weight of the menu item. A lower number means it will be placed higher."] = 10.0
    name: Annotated[str, "Name of the item"] = "Unnamed item"
    description: Annotated[str, "The item's description, what it's supposed to do"] = "Coders fix this"

    @abstractmethod
    def run(self) -> None:
        """What it says on the tin. Interactively runs what the item's supposed to do. What else did you expect?"""

    # The methods below are very useful for sorting
    def __str__(self) -> str:
        return f"{self.name}: {self.description}"

    def __lt__(self, other) -> bool:
        return self.weight < other.weight and self.description < other.name

    def __gt__(self, other) -> bool:
        return self.weight > other.weight and self.description > other.name

    def __eq__(self, other) -> bool:
        return self.weight == other.weight and self.description == other.name

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)



import requests as req

from .ban_types import BanData
from .constants import CENTCOM_API_URL
from ..constants import USER_AGENT
from ..byond import canonicalize


def get_one(key: str):
    """Gets bans for a single key"""
    ckey = canonicalize(key)
    resp = req.get(CENTCOM_API_URL.format(ckey=ckey), timeout=10, headers={"User-Agent": USER_AGENT})
    return BanData.from_response(resp)



from typing import Annotated, Iterable

from .abstract import LogDownloader
from .constants import DEFAULT_FILES, DEFAULT_ROUND_OUTPUT_PATH
from ..scrubby.RoundController import round_ids_to_round_data


class RoundLogDownloader(LogDownloader):
    """Downloads a span of rounds"""
    lbound: Annotated[int, "Left boundary"]
    rbound: Annotated[int, "Right boundary"]

    def __init__(self, start_round: int, end_round: int, output_path: str = None) -> None:
        self.lbound = min(start_round, end_round)
        self.rbound = max(start_round, end_round)
        self.output_path = output_path or DEFAULT_ROUND_OUTPUT_PATH.format(start=self.lbound, end=self.rbound)

    async def update_round_list(self) -> None:
        def round_list_generator():
            i = self.lbound
            while i <= self.rbound:
                yield i
                i += 1
        async for round_data in round_ids_to_round_data(round_list_generator()):
            self.rounds.append(round_data)

    def filter_lines(self, logs: list[bytes]) -> Iterable[bytes]:
        return logs

    @staticmethod
    def interactive() -> LogDownloader:
        while True:
            try:
                start = int(input("First round: "))
                break
            except ValueError:
                print("Could not parse that as a number, please try again")
        while True:
            try:
                end = int(input("Last round (inclusive): "))
                break
            except ValueError:
                print("Could not parse that as a number, please try again")
        output_path = input(f"Where should I write the file? [{DEFAULT_ROUND_OUTPUT_PATH}] ")
        downloader = RoundLogDownloader(start, end, output_path)
        print("Which files do you want to download?")
        print("(separate the files with a comma, like so: attack.txt,game.txt,pda.txt)")
        file_list = [x.strip() for x in input(f"[{','.join(DEFAULT_FILES)}] ").split(',') if x.strip()]
        if file_list:
            downloader.files = file_list
        downloader.try_authenticate_interactive()
        return downloader



"""
This file contains the menu items that should be loaded into the main menu.
All of them must inherit from MenuItem
"""
# pylint: disable=missing-class-docstring,too-few-public-methods
import asyncio
import traceback
import sys

from colorama import Style, Fore

from .menu_item import MenuItem
from .log_downloader import CkeyLogDownloader, RoundLogDownloader

try:
    from .slur_detector import SlurDetector
except FileNotFoundError:
    # Bit of a hack but it does the job
    print(traceback.format_exc().replace("FileNotFoundError:", f"{Fore.RED}FileNotFoundError:") + Fore.RESET)
    print(f"{Style.DIM}Press return to exit...{Style.RESET_ALL}", end='')
    input()
    sys.exit(1)


class CkeySingleItem(MenuItem):
    weight = 1
    name = "ckey log downloader"
    description = "Download someone's say history (and more!)"

    def run(self):
        downloader = CkeyLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())


class SlurDetectorSingleItem(MenuItem):
    name = "slur detector"
    description = "Run slur detection on a file"

    def run(self):
        from .slur_detector import main  # pylint: disable=import-outside-toplevel
        main()


class CkeyAndSlurItem(MenuItem):
    weight = 2
    name = "ckey log slur detector"
    description = "Run slur detection on someone's say logs"

    def run(self):
        downloader = CkeyLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())
        print()  # Newline
        slurs = SlurDetector.from_file(downloader.output_path)
        slurs.print_results()


class RoundSingleItem(MenuItem):
    weight = 3
    name = "round log downloader"
    description = "Download logs from a range of rouds"

    def run(self):
        downloader = RoundLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())


class RoundAndSlurItem(MenuItem):
    name = "round slur detector"
    description = "Run slur detector on a range of rouds"

    def run(self):
        downloader = RoundLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())
        print()  # Newline
        slurs = SlurDetector.from_file(downloader.output_path)
        slurs.print_results()


class CentComItem(MenuItem):
    name = "CentCom"
    description = "Search the CentCom ban database for ckeys"

    def run(self):
        from .centcom import main  # pylint: disable=import-outside-toplevel
        main()


class UserExistsItem(MenuItem):
    name = "BYOND user exists"
    description = "Check if users on a list exist or not"

    def run(self):
        from .byond import main  # pylint: disable=import-outside-toplevel
        main()


class TokenTestServiceItem(MenuItem):
    name = "Token test service"
    description = "The one-stop shop for TG13 token testing"

    def run(self):
        from .auth import main  # pylint: disable=import-outside-toplevel
        main()



"""Wrapper for the CKeyController on Scrubby"""
# pylint: disable=invalid-name
from aiohttp import ClientSession

from .round_data import RoundData
from .constants import PLAYER_ROUNDS_URL
from ..constants import USER_AGENT


class ScrubbyException(Exception):
    """Wuh oh, scrubby died"""


async def GetReceipts(ckey: str, number_of_rounds: int, only_played: bool = False) -> list[RoundData]:
    """Calls the scrubby API and retrieves the specified number of rounds"""
    data = {
        "ckey": ckey,
        "startingRound": 999999,  # that's how scrubby does it, sue Bobbah, not me
        "limit": number_of_rounds
    }
    # ckey is specified twice, but it seems like the url ckey does not matter at all
    # https://github.com/bobbahbrown/ScrubbyWebPublic/blob/d71ad368e156f56524bf7ec323685ca29af35baa/Controllers/CKeyController.cs#L78
    async with ClientSession(headers={"User-Agent": USER_AGENT}) as session:
        resp = await session.post(PLAYER_ROUNDS_URL.format(ckey=ckey), json=data)
        if not resp.ok:
            raise ScrubbyException("Scrubby errored with code " + str(resp.status))
        if await resp.read() == b"[]":
            raise ScrubbyException("CKEY could not be found")
        if not only_played:
            return await RoundData.from_scrubby_response_async(resp)

        played_in = []
        while True:
            rounds = await RoundData.from_scrubby_response_async(resp)
            if not rounds:
                return played_in
            for round_data in rounds:
                if round_data.playedInRound:
                    played_in.append(round_data)
                if len(played_in) == number_of_rounds:
                    return played_in
            data["startingRound"] = rounds[-1].roundID
            resp = await session.post(PLAYER_ROUNDS_URL.format(ckey=ckey), json=data)



from __future__ import annotations
from typing import Annotated, Iterable

from colorama import Fore

from .slur_file import SLURS
from .word_detection import detect_word


class SlurDetector:
    """Opens the file and scans for slurs. Results stored in tally"""

    tally: Annotated[dict[str, int], "A dictionary with strings (slurs) as keys, and their tally as key (int)"]
    slur_lines: Annotated[list[tuple[str, str]], "Stores all unmodified detected lines with the slur's position"]

    def __init__(self, text: Iterable[str]) -> None:
        self.reset_tally()
        self.scan_text(text)

    def reset_tally(self):
        """Resets the tally to 0"""
        self.tally = {}
        for slur in SLURS:
            self.tally[slur] = 0
        self.slur_lines = []

    def scan_text(self, text: Iterable[str]) -> None:
        """Scans the text. Automatically called in __init__"""
        for line in text:
            self.process_line(line)

    def process_line(self, text: str) -> None:
        """Processes one line and detects possible slurs"""
        for slur in SLURS:
            if detect_word(slur, text):
                self.slur_lines.append((text.strip(), slur))
                self.tally[slur] += 1

    def print_tally(self):
        """Prints the slurs according to the tally"""
        print("\nSlurs:")
        none = True
        for key, value in self.tally.items():
            if value:
                none = False
                print(f"{key}\t{value}")
        if none:
            print(f"{Fore.GREEN}None!{Fore.RESET}")

    def print_results(self):
        """Prints the results to the console"""
        self.print_slur_lines()
        self.print_tally()

    def print_slur_lines(self):
        """Prints all of the lines which have slurs in them"""
        print(f"{Fore.YELLOW}Lines with detected slurs:{Fore.RESET}")
        for slur_line, slur in self.slur_lines:
            print(slur_line.replace(slur, f"{Fore.RED}{slur}{Fore.RESET}"))

    @staticmethod
    def from_file(target_file: str) -> SlurDetector:
        """Opens the path and scans the contents with a new SlurDetector"""
        with open(target_file, "r", encoding="utf-8") as file:
            return SlurDetector(file.readlines())



def detect_word(word: str, text: str) -> bool:
    """Detects if the specified word is in the text"""
    # TODO: improve functionality
    # TODO: add Snowflake stemmer maybe?
    return word in text



"""Abstract implementation of a log downloader"""
from __future__ import annotations
import asyncio
from abc import ABC, abstractmethod
from typing import Generator, Iterable, Annotated, Union
from itertools import product

from aiohttp import ClientSession
from colorama import Fore, Style
from dateutil.parser import isoparse
from tqdm.asyncio import tqdm

from .constants import DEFAULT_OUTPUT_PATH, GAME_TXT_URL, GAME_TXT_ADMIN_URL, DEFAULT_FILES
from ..constants import USER_AGENT, POSITIVE_RESPONSES, NEGATIVE_RESPONSES
from ..scrubby import RoundData
from ..auth import is_authenticated, create_from_token, interactive as tgauth_interactive, get_auth_headers, seconds_left


class LogDownloader(ABC):
    """Log downloader object. For downloading logs.
    Either pass the arguments in the constructor or call `interactive()`"""

    user_agent: Annotated[str, "User agent so people know who keeps spamming requests (and for raw logs)"] = USER_AGENT
    output_path: Annotated[str, "Where should we write the file to?"] = DEFAULT_OUTPUT_PATH
    rounds: Annotated[list[RoundData], "The list of rounds to download"] = []
    files: Annotated[list[str], "Which files do we want to dowload?"] = DEFAULT_FILES.copy()

    def authenticate(self, token: str, override_old: bool) -> bool:
        """Tries to authenticate against the TG forums"""
        if is_authenticated():
            return True
        return create_from_token(token=token, override_old=override_old)

    def try_authenticate_interactive(self) -> bool:
        """Tries to authenticate against the TG forums interactively"""
        if is_authenticated() and seconds_left() < 30:
            print(f"{Fore.YELLOW}WARNING{Fore.RESET}: token has less than 30 seconds left")
            print(f"Refresh the token? [{Style.BRIGHT}Y{Style.NORMAL}/n] ", end='')
            if input().lower() not in NEGATIVE_RESPONSES:
                while not create_from_token(input("Token: ").strip(), True):
                    pass  # There's a print if Passport errors on creation. Terrible code. Future me will fix it.
                return True
        print(f"Would you like to access raw logs? [y/{Style.BRIGHT}N{Style.NORMAL}] ", end='')
        if input().lower() not in POSITIVE_RESPONSES:
            return False
        return tgauth_interactive()

    @abstractmethod
    async def update_round_list(self) -> None:  # Not the best way of doing it but I can't be bothered right now
        """Generates a list of rounds and saves it to self.rounds"""

    def get_log_links(self) -> Iterable[str]:
        """Gets the links of logs we want to download"""
        if is_authenticated():
            url = GAME_TXT_ADMIN_URL
            self.files = [file.replace(".txt", ".log") for file in self.files]
        else:
            url = GAME_TXT_URL
            self.files = [file.replace(".log", ".txt") for file in self.files]

        for round_data, file_name in product(self.rounds, self.files):
            round_data.timestamp = isoparse(round_data.timestamp)
            yield round_data, url.format(
                server=round_data.server.lower().replace('bagil', 'basil').replace(' ', '-'),
                year=str(round_data.timestamp.year),
                month=f"{round_data.timestamp.month:02d}",
                day=f"{round_data.timestamp.day:02d}",
                round_id=round_data.roundID,
                file_name=file_name
            )

    @abstractmethod
    def filter_lines(self, logs: list[bytes]) -> Iterable[bytes]:
        """Filters lines from a log file, returning only the ones we want"""

    async def get_logs_async(self) -> Generator[tuple[RoundData, Union[list[bytes], None]], None, None]:
        """This is a generator that yields a tuple of the `RoundData` and list of round logs, for all rounds in `rounds`

        if `output_bytes` is true, the function will instead yield `bytes` instead of `str`

        On 404, the list will be None instead"""
        headers = {"User-Agent": self.user_agent}
        if is_authenticated():
            headers.update(get_auth_headers())
        async with ClientSession(headers=headers) as session:
            tasks = []

            async def fetch(round_data: RoundData, link: str):
                # Edge case warning: if we go beyond the year 2017 or so, the logs path changes.
                # I don't expect anyone to go that far so I won't be doing anything about it
                async with session.get(link) as rsp:
                    if not rsp.ok:
                        return round_data, None
                    return round_data, await rsp.read()

            for round_data, link in self.get_log_links():
                tasks.append(asyncio.ensure_future(fetch(round_data, link)))

            # This could be out of order but we don't really care, it's not important
            for task in tasks:
                response: bytes
                round_data, response = await task
                if not response:
                    yield round_data, None
                else:
                    yield round_data, response.replace(b'\r', b'').split(b'\n')

            await asyncio.gather(*tasks)

    @staticmethod
    def format_line_bytes(line: bytes, round_data: RoundData) -> list[str]:
        """Takes the raw line and formats it to `{server_name} {round_id} | {unmodified line}`"""
        return round_data.server.encode("utf-8") + b" " + str(round_data.roundID).encode("utf-8") + b" | " + line + b"\n"

    async def process_and_write(self, output_path: str = None):
        """Processes the data, downloads the logs and saves them to a file"""
        output_path = output_path or self.output_path
        if not self.rounds:
            await self.update_round_list()
        with open(output_path, 'wb') as file:
            pbar = tqdm(self.get_logs_async(), total=len(self.rounds)*len(self.files))
            async for round_data, logs in pbar:
                # Type hints
                round_data: RoundData
                logs: list[bytes]

                pbar.set_description(f"Getting ID {round_data.roundID} on {round_data.server}")
                if not logs:
                    pbar.clear()
                    print(f"{Fore.YELLOW}WARNING:{Fore.RESET} Could not get a file from round " +
                          f"{round_data.roundID} on {round_data.server}")
                    pbar.display()
                    continue
                if round_data.roundStartSuicide:
                    pbar.clear()
                    print(f"{Fore.MAGENTA}WARNING:{Fore.RESET} round start suicide " +
                          f"in round {round_data.roundID} on {round_data.server}")
                    pbar.display()
                for line in self.filter_lines(logs):
                    file.write(self.format_line_bytes(line, round_data))

    @staticmethod
    @abstractmethod
    def interactive() -> LogDownloader:
        """Interactively set variables"""



from string import ascii_lowercase, digits
from urllib.parse import quote
from typing import Generator
import asyncio
import sys

import requests as req
from aiohttp import ClientSession

from .constants import BYOND_MEMBERS_URL
from ..constants import USER_AGENT


def canonicalize(key: str) -> str:
    """Turns a user's key into canonical form (ckey)"""
    return ''.join([letter for letter in key.lower() if letter in ascii_lowercase + digits + '@'])


def user_exists(key: str) -> bool:
    """Queries the BYOND website and figures out if a user is real"""
    ckey = canonicalize(key)
    # canonicalize should make it url safe but just in case let's also use quote
    resp = req.get(BYOND_MEMBERS_URL.format(ckey=quote(ckey)), timeout=10, headers={"User-Agent": USER_AGENT})
    # The issue here is that when a user doesn't exist, the website does NOT redirect
    # nor does it return a 404. It just goes on as usual. The easiest way is doing this
    # and hoping it doesn't break one day.
    if not resp.ok:
        raise req.ConnectionError(f"Got {resp.status_code} instead of 200")
    if "not found" not in resp.text and '<span>Joined: <span class="info_text">' in resp.text:
        return True
    return False


async def user_exists_many(keys: list[str]) -> Generator[bool, None, None]:
    """Queries the BYOND website and figures out if the users are real"""
    tasks = []
    async with ClientSession(headers={"User-Agent": USER_AGENT}) as session:
        async def fetch_one(ckey: str):
            async with session.get(BYOND_MEMBERS_URL.format(ckey=ckey)) as rsp:
                if not rsp.ok:
                    print(f"Request to {BYOND_MEMBERS_URL} returned status {rsp.status} instead of ok", file=sys.stderr)
                return await rsp.read()
        for key in keys:
            ckey = canonicalize(key)
            # canonicalize should make it url safe but just in case let's also use quote
            tasks.append(asyncio.ensure_future(fetch_one(quote(ckey))))

        # The issue here is that when a user doesn't exist, the website does NOT redirect
        # nor does it return a 404. It just goes on as usual. The easiest way is doing this
        # and hoping it doesn't break one day.
        for task in tasks:
            resp = await task
            yield (b"not found" not in resp and b'<span>Joined: <span class="info_text">' in resp)
        asyncio.gather(tasks)



# pylint: disable=invalid-name
from typing import Generator, Iterable
import asyncio
import sys

import requests as req
from aiohttp import ClientSession

from .constants import ROUND_SOURCE_URL
from . import RoundData
from ..constants import USER_AGENT


def get_round_source_url(round_id: int) -> str:
    """Gets the source url from a round ID"""
    return req.get(ROUND_SOURCE_URL.format(round_id=str(round_id)), timeout=10,
                   headers={"User-Agent": USER_AGENT}).json()['baseURL']


async def get_multiple_round_json(round_ids: Iterable[str]) -> Generator[dict, None, None]:
    """Queries scrubby and retrieves round info JSONs"""
    tasks = []
    async with ClientSession(headers={"User-Agent": USER_AGENT}) as session:
        async def fetch_one(round_id: str):
            async with session.get(ROUND_SOURCE_URL.format(round_id=str(round_id))) as resp:
                if not resp.ok:
                    print(f"Request {round_id} returned status {resp.status} instead of ok", file=sys.stderr)
                    return None
                return await resp.json()
        for round_id in round_ids:
            round_id = round_id if round_id is str else str(round_id)
            tasks.append(asyncio.ensure_future(fetch_one(round_id)))

        for task in tasks:
            resp = await task
            yield resp
        await asyncio.gather(*tasks)


async def get_multiple_round_source_urls(round_ids: Iterable[str]) -> Generator[bool, None, None]:
    """Gets source URLs for multiple rounds"""
    async for rnd in get_multiple_round_json(round_ids=round_ids):
        if not rnd:
            continue
        yield rnd['baseURL']


async def round_ids_to_round_data(round_ids: Iterable[str]) -> Generator[RoundData, None, None]:
    """Contructs round data objects from round IDs"""
    async for rnd in get_multiple_round_json(round_ids=round_ids):
        if not rnd:
            continue
        round_info = rnd['currentRound']
        yield RoundData.should_not_be_used_this_way(
            round_id=round_info['id'],
            timestamp=round_info['startTime'],
            server=round_info['server']
        )



from colorama import init as colorama_init

from .__version__ import __version__

colorama_init()

VERSION = __version__

__all__ = [
    'slur_detector',
    'centcom',
    'byond',
    'log_downloader'
]

del colorama_init



SCRUBBY_URL = "https://scrubby.melonmesa.com/"
PLAYER_ROUNDS_URL = SCRUBBY_URL + "ckey/{ckey}/receipts"
ROUND_SOURCE_URL = SCRUBBY_URL + "round/{round_id}?raw=true"
# Locked behind [Authorize(Roles = "Developer,In-Game Admin,BetaTester")]
# https://github.com/bobbahbrown/ScrubbyWebPublic/blob/d71ad368e156f56524bf7ec323685ca29af35baa/Controllers/InvestigateController.cs#L12
# COMMON_ROUNDS_URL = SCRUBBY_URL + "api/roundsforckeys"



import sys

from .slur_detector import SlurDetector


def main():
    """Main"""
    if len(sys.argv) == 2:
        file_name: str = sys.argv[1]
    else:
        file_name = input("File name: ")
    slur_search = SlurDetector.from_file(file_name)
    slur_search.print_results()


if __name__ == "__main__":
    main()



from ss13_tools.centcom.ban import get_one


def main():
    """Main"""
    try:
        print("Paste ckeys to search for, one per line (press CTRL + C to stop)\n")
        while True:
            key = input()
            ban_data = get_one(key)

            print(f"{key}:\n")
            if len(ban_data) == 0:
                print("No data")
            for ban in ban_data:
                print(ban.sourceName, ban.sourceRoleplayLevel)
                print(f"Banned on {ban.bannedOn} by {ban.bannedBy}")
                print("With reason:", ban.reason)
                print("Expires:", ban.expires or "never", "and is", "active" if ban.active else "not active")
            print("=========================\n")
    except KeyboardInterrupt:
        print("Bye!")


if __name__ == "__main__":
    main()



import sys
import asyncio

from colorama import Fore

from .constants import DEFAULT_ONLY_PLAYED, DEFAULT_CKEY_OUTPUT_PATH, DEFAULT_NUMBER_OF_ROUNDS
from . import CkeyLogDownloader, RoundLogDownloader
from .abstract import LogDownloader


def make_ckey_downloader() -> LogDownloader:
    """Tries to construct a ckey downloader from args"""
    ckey = sys.argv[2]
    number_of_rounds = int(sys.argv[3])
    output_path = DEFAULT_CKEY_OUTPUT_PATH.format(ckey=ckey)
    if len(sys.argv) > 4:
        output_path = sys.argv[4]
    only_played = DEFAULT_ONLY_PLAYED
    if len(sys.argv) > 5:
        only_played = sys.argv[5]
        raise NotImplementedError()
    return CkeyLogDownloader(ckey, number_of_rounds, output_path, only_played)


def make_round_id_downloader() -> LogDownloader:
    """Tries to construct a round id downloader from args"""
    file_path = sys.argv[4] if len(sys.argv) == 4 else f"{sys.argv[1]}-{sys.argv[2]}"
    return RoundLogDownloader(int(sys.argv[2]), int(sys.argv[3]), file_path)


def print_help():
    """Prints the help text"""
    print(f"{Fore.YELLOW}Unknown number of command line arguments{Fore.RESET}")
    print(f"{Fore.GREEN}USAGE{Fore.RESET}: {sys.argv[0]} ckey <ckey> [number_of_rounds={DEFAULT_NUMBER_OF_ROUNDS}]" +
          "[output_path={ckey}.txt] [only_played=false]")
    print(f"{Fore.GREEN}USAGE{Fore.RESET}: {sys.argv[0]} round <start_round> <end_round> [output_path]")
    print("<> are required, [] are optional, = means a default value. If you provide an optional, you have to also " +
          "provide all optionals before it")


# Do argparse
if 2 < len(sys.argv) < 7:
    if sys.argv[1] == "round":
        downloader = make_round_id_downloader()
    elif sys.argv[1] == "ckey":
        downloader = make_ckey_downloader()
    else:
        print_help()
        sys.exit(1)
elif len(sys.argv) != 1:
    print_help()
    sys.exit(1)
else:
    downloader = CkeyLogDownloader()
    downloader.interactive()
    asyncio.run(downloader.process_and_write())



from colorama import Fore, Style

from .key_tools import user_exists


def main():
    """Main"""
    try:
        print("Paste ckeys to search for, one per line (press CTRL + C to stop)\n")
        while True:
            key = input()
            exists = user_exists(key)
            print(Fore.GREEN if exists else Fore.RED, end='')
            print(f"{Style.BRIGHT}{key}{Style.NORMAL}", "exists" if exists else "does not exist", Fore.RESET)
    except KeyboardInterrupt:
        print("Bye!")


if __name__ == "__main__":
    main()



import os

from colorama import Fore

from .constants import SLURS_FILE

if not os.path.exists(SLURS_FILE):
    with open(SLURS_FILE, "w", encoding='utf-8') as file:
        file.write("### Slurs file\n" +
                   "### One slur per line, # to ignore the line\n" +
                   "### The program will also skip empty lines\n\n")
    raise FileNotFoundError("Slurs file does not exist. Creating it for you. Please add some words to it.")

SLURS = []  # Should not be modified after filling up once
with open(SLURS_FILE, "r", encoding="utf-8") as file:
    for line in file.readlines():
        if line.strip() and not line.startswith('#'):
            SLURS.append(line.split('#', 1)[0].strip())

if not SLURS:
    print(f"{Fore.YELLOW}WARNING:{Fore.RESET} No slur entries detected! Please open the slurs file and add some.")

del SLURS_FILE  # Del as you probably don't want this constant anyway



import sys

from .__version__ import __version__

USER_AGENT = f"ss13tools/{__version__} ({sys.platform}) Python {sys.version}"
POSITIVE_RESPONSES = ['y', 'yes', 'true', '1']
NEGATIVE_RESPONSES = ['n', 'no', 'false', '0']



import sys
import asyncio

from colorama import init as colorama_init

from .ckey import CkeyLogDownloader
from .round import RoundLogDownloader
from ..scrubby import RoundData

colorama_init()

if sys.platform == "win32":
    # This fixes a lot of runtime errors.
    # It's supposed to be fixed but oh well.
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

__all__ = [
    'CkeyLogDownloader',
    'RoundLogDownloader',
    'RoundData'
]

del colorama_init



# pylint: disable=invalid-name
# Enable type hinting for static methods
from __future__ import annotations
import json
from typing import Optional, Annotated, Any
from dataclasses import dataclass

from aiohttp import ClientResponse


@dataclass
class RoundData:  # pylint: disable=too-many-instance-attributes
    """Represents round data from Scrubby"""
    roundID: int
    job: Optional[str]
    timestamp: Annotated[str, "ISO 8601, YYYY-MM-DDThh:mm:ss.ffffZ"]
    connectedTime: Annotated[str, "hh:mm:ss.fffffff"]
    roundStartPlayer: bool
    playedInRound: bool
    antagonist: bool
    roundStartSuicide: bool
    isSecurity: bool
    firstSuicide: bool
    firstSuicideEvidence: Optional[Any]
    name: Optional[str]
    server: str

    @staticmethod
    async def from_scrubby_response_async(r: ClientResponse) -> list[RoundData]:
        """Converts a Scrubby JSON response directly to a Python RoundData object"""
        return json.loads(await r.text(), object_hook=lambda d: RoundData(**d))

    # TODO: someone needs to fix this, it's a horrible way to do it
    @staticmethod
    def should_not_be_used_this_way(round_id: int, timestamp: str, server: str) -> RoundData:
        # This needs to be reworked. heavily. log_downloader is still a mess. Better than before, still a mess.
        # self.rounds is still used, that needs to be reworked since it relies on RoundData (this thing)
        # Right now I'm not in the mood, so I'll push it down for future me.
        """Those lazy devs always up to no good"""
        return RoundData(
            round_id,
            None,
            timestamp,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            server
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ss13_tools',
 'ss13_tools.auth',
 'ss13_tools.byond',
 'ss13_tools.centcom',
 'ss13_tools.log_downloader',
 'ss13_tools.scrubby',
 'ss13_tools.slur_detector']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0',
 'colorama>=0.4.6,<0.5.0',
 'pycryptodome>=3.17,<4.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.28.1,<3.0.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'ss13-tools',
    'version': '2.1.0',
    'description': 'Python toolchain for SS13',
    'long_description': "# SS13 tools\n\n[![PyLint](https://github.com/RigglePrime/SS13-tools/actions/workflows/pylint.yml/badge.svg)](https://github.com/RigglePrime/SS13-tools/actions/workflows/pylint.yml)\n[![Flake8](https://github.com/RigglePrime/SS13-tools/actions/workflows/flake8-lint.yml/badge.svg)](https://github.com/RigglePrime/SS13-tools/actions/workflows/flake8-lint.yml)\n[![Build](https://github.com/RigglePrime/SS13-tools/actions/workflows/publish.yml/badge.svg)](https://github.com/RigglePrime/SS13-tools/actions/workflows/publish.yml)\n\nA set of tools to help with automating tasks for the /tg/ branch of SS13.\n\nMade for Python 3.9+\n\n## How to run\n\n`pip install ss13-tools; python -m ss13_tools` (pip3 on Linux) or dowload the executable [here](https://github.com/RigglePrime/SS13-tools/releases/latest)\n\n## Contents\n\n- auth (TODO)\n- byond: tools for working with BYOND, such as checking if a ckey exists\n- [centcom](https://centcom.melonmesa.com/)\n- log_downloader: downloads logs from tg's [parsed logs](https://tgstation13.org/parsed-logs)\n- scrubby: scrubby tools\n- slur_detector: what it says on the tin\n\n## Downloading\n\nFor compiled versions, see [this](https://github.com/RigglePrime/admin-tools/releases) link\n\nNow also available on [PyPi](https://pypi.org/project/ss13-tools/)!\n",
    'author': 'RigglePrime',
    'author_email': '27156122+RigglePrime@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/RigglePrime/SS13-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.1,<3.12',
}


setup(**setup_kwargs)



BASE_URL = "https://tgstation13.org/"
LOGS_URL = BASE_URL + "parsed-logs/"
RAW_LOGS_URL = BASE_URL + "raw-logs/"
AUTH_TEST_URL = RAW_LOGS_URL
TOKEN_URL = BASE_URL + "get_rawlogs_token.php"
PASSPORT_URL = BASE_URL + "get_rawlogs_passport.php"
PASSPORT_FILE_LOCATION = "./"
PASSPORT_FILE_NAME = "passport.auth"
PASSPORT_FILE_HEADER = b"TG\x03\x04"



#!/env/python3

import traceback
from random import choice
import sys

from colorama import Fore, Style

from ss13_tools.menu import MenuItem  # Absolute import because of pyinstaller


colour = choice([Fore.BLUE, Fore.CYAN, Fore.GREEN, Fore.LIGHTBLUE_EX, Fore.LIGHTCYAN_EX,
                Fore.LIGHTGREEN_EX, Fore.LIGHTMAGENTA_EX, Fore.LIGHTRED_EX, Fore.LIGHTWHITE_EX,
                Fore.LIGHTYELLOW_EX, Fore.MAGENTA, Fore.RED, Fore.WHITE, Fore.YELLOW])

print(f"Welcome to {colour}ss13-tools{Fore.RESET}! What would you like to do?")
classes = {str(i): x for i, x in enumerate(sorted(x() for x in MenuItem.__subclasses__()), 1)}

for key, value in classes.items():
    print(f"{Fore.GREEN}{key}{Fore.RESET}: {value.description}")

try:
    choice = input()  # Colorama and input don't mix well :/
    classes[choice].run()
except KeyError:
    print("Invalid choice")
except KeyboardInterrupt:
    sys.exit(0)
except Exception:  # pylint: disable=broad-except
    traceback.print_exc()

print(f"\n{Style.DIM}Press return to exit...{Style.RESET_ALL}", end='')
input()



# pylint: disable=invalid-name,too-many-instance-attributes
from __future__ import annotations  # Required so we can reference BanData in its own staticmethod
from typing import Optional
from enum import Enum
from dataclasses import dataclass
import json

from requests import Response


class RoleplayLevel(Enum):
    """The level of roleplay"""
    Unknown = 0
    Low = 1
    Medium = 2
    High = 3


class BanType(Enum):
    """What type of ban is it?"""
    Server = 0
    Job = 1


@dataclass(repr=True, frozen=True)
class BanData:
    """Object that stores ban info"""
    id: int = -1
    sourceID: int = -1
    sourceName: Optional[str] = None
    sourceRoleplayLevel: RoleplayLevel = RoleplayLevel.Unknown
    type: BanType = BanType.Server
    cKey: Optional[str] = None
    bannedOn: str = "ERROR"
    bannedBy: Optional[str] = None
    reason: Optional[str] = None
    expires: Optional[str] = None
    unbannedBy: Optional[str] = None
    banID: Optional[str] = None
    jobs: Optional[list[str]] = None
    banAttributes: Optional[list[str]] = None
    active: bool = False

    @staticmethod
    def from_response(r: Response) -> list[BanData]:
        """Get the JSON of this requests response and map it to the object"""
        return r.json(object_hook=lambda d: BanData(**d))

    @staticmethod
    def from_json_string(r: str) -> list[BanData]:
        """Map JSON values to this object and return it"""
        return json.loads(r, object_hook=lambda d: BanData(**d))



"""
Auth module

Provides methods to authenticate against raw logs
"""
from __future__ import annotations

# Grrr globals but right now I don't care, refactor later
# Maybe if another station needs to add their own auth

import time
import os
import platform
import locale
import pickle
import struct
from datetime import datetime
from typing import Optional

from colorama import Fore, Style
from Crypto.Hash import SHA256
from Crypto.Cipher import AES
import requests as req

from ..constants import USER_AGENT
from .constants import PASSPORT_FILE_LOCATION, PASSPORT_FILE_NAME, PASSPORT_FILE_HEADER, AUTH_TEST_URL,\
                       TOKEN_URL, PASSPORT_URL


PASSPORT: Passport = None


class Passport():
    """Passport class"""

    rawlogs_passport: str
    expires_at: datetime
    current_server_time: datetime

    def __init__(self, token: str) -> None:
        # Request a new passport to be issued
        passport_response = req.get(PASSPORT_URL,
                                    headers={"User-Agent": USER_AGENT, "Authorization": f"Bearer {token}"},
                                    timeout=10)
        if not passport_response.ok:
            return

        # Shush I'm lazy I will do it this way for now
        passport_response = passport_response.json()
        if (err := passport_response.get("error")):
            print(err)
            return
        self.rawlogs_passport = passport_response["rawlogs_passport"]
        self.expires_at = datetime.fromtimestamp(passport_response["expires_at"])
        self.current_server_time = datetime.fromtimestamp(passport_response["current_server_time"])

    def seconds_left(self) -> float:
        """The number of seconds this token has left before impending death. Now including fractions!"""
        return self.expires_at.timestamp() - datetime.utcnow().timestamp()

    def test(self) -> bool:
        """Tests the passport for validity"""
        return bool(self.rawlogs_passport) and\
            self.expires_at > datetime.utcnow() and\
            req.get(AUTH_TEST_URL,
                    headers={"User-Agent": USER_AGENT, "Authorization": f"Bearer {self.rawlogs_passport}"},
                    timeout=10).ok

    def save_to_file(self, path: str = PASSPORT_FILE_LOCATION + PASSPORT_FILE_NAME) -> None:
        """Saves the passport to a file"""
        with open(path, 'ab+') as file:
            file.seek(0, 0)  # Append puts us at the end of the file
            header = file.read(len(PASSPORT_FILE_HEADER))  # Try read the header
            if file.tell() and header and header != PASSPORT_FILE_HEADER:
                raise FileExistsError(f"The file at {path} already exists and is not a passport file")

            cipher = AES.new(generate_key(), AES.MODE_GCM)
            passport_pickle = pickle.dumps(self)
            passport_pickle, mac = cipher.encrypt_and_digest(passport_pickle)

            file.seek(0, 0)
            file.truncate()
            file.write(PASSPORT_FILE_HEADER)
            file.write(struct.pack("<B", len(cipher.nonce)))
            file.write(cipher.nonce)
            file.write(struct.pack("<I", len(passport_pickle)))
            file.write(passport_pickle)
            file.write(struct.pack("<B", len(mac)))
            file.write(mac)

    @staticmethod
    def load_from_file(path: str = PASSPORT_FILE_LOCATION + PASSPORT_FILE_NAME) -> Optional[Passport]:
        """Reads a passport from a file"""
        with open(path, 'rb') as file:
            header = file.read(len(PASSPORT_FILE_HEADER))  # Try read the header
            # file.tell tells us if we moved in the file at all, check header if it's empty, and then validate it
            if file.tell() and header and header != PASSPORT_FILE_HEADER:
                raise FileNotFoundError(f"The file at {path} is not a passport file")

            nonce_len = struct.unpack('<B', file.read(1))[0]  # Yes, yes, <B and B are the same, shush
            nonce = file.read(nonce_len)
            cipher = AES.new(generate_key(), AES.MODE_GCM, nonce=nonce)
            pickle_len = struct.unpack('<I', file.read(4))[0]
            passport_file = file.read(pickle_len)
            mac_len = struct.unpack('<B', file.read(1))[0]
            mac = file.read(mac_len)
            try:
                passport_file = cipher.decrypt_and_verify(passport_file, mac)
            except ValueError:  # Wuh oh, mac doesn't match! Someone tampered with our file!
                return None  # NEVER load a pickle you don't trust
            return pickle.loads(passport_file)


def generate_key() -> bytes:
    """
    Generates the 256-bit encryption/decryption key.
    This is NOT meant to be safe, and is NOT a replacement
    for a safe, user-defined password, just safe enough so
    our user doesn't accidentally share their passport.
    Or if their files get leaked, the passport is protected.

    Passports are by their very nature short lived. Because
    saving it is insecure, tokens aren't stored next to them.
    (tokens have a longer lifespan)

    Hopefully if someone asks our user for all of this info
    they will get extremely suspicious. What I made is probably
    safer than most users' passwords anyway.
    """
    bits, linkage = platform.architecture()
    loc, encoding = locale.getlocale()
    key: str = platform.node() + os.getlogin() + bits + linkage + time.tzname[0] + loc + encoding
    sha = SHA256.new()
    sha.update(key.encode())
    return sha.digest()


def save_passport() -> None:
    """Saves the provided passport"""
    PASSPORT.save_to_file()


def load_passport() -> None:
    """Loads the passport into the auth module"""
    global PASSPORT  # pylint: disable=global-statement
    try:
        PASSPORT = Passport.load_from_file()
    except FileNotFoundError:
        pass


def create_from_token(token: str, override_old: bool = False) -> bool:
    """Creates a new passport from a token. Tests the token before setting it, retuning if the token is valid or not

    If override_old is not true, the new token won't be used if the old one is still valid"""
    if is_authenticated() and not override_old:
        return True
    global PASSPORT  # pylint: disable=global-statement
    new_passport = Passport(token=token)
    if not new_passport.test():
        return False
    PASSPORT = new_passport
    return True


def interactive():
    """Interactively authenticate. Calls functions to load and test the passport before asking the user"""
    global PASSPORT  # pylint: disable=global-statement
    print("Authenticating...", end='')
    if is_authenticated():
        print(Fore.GREEN, "passport valid", Fore.RESET)
        valid_for = PASSPORT.expires_at - datetime.utcnow()
        print(f"The token has {valid_for} left.")
        return True
    print(Fore.RED, "not authenticated!", Fore.RESET)
    print(f"Please go to {Style.BRIGHT}{TOKEN_URL}{Style.NORMAL} and obtain a token. The token will NOT be saved " +
          f"in this software, so please store it somewhere safe.\nWith it, anyone could access raw logs, {Fore.CYAN}" +
          f"treat it as you would treat a password{Fore.RESET}! {Fore.RED}If you accidentally leak this, change your " +
          f"password immediately{Fore.RESET} to invalidate it.")
    print(f"If you are having trouble copy-pasting the link, select it with your mouse and {Fore.CYAN}right click" +
          f"{Fore.RESET}. That should copy it.")
    print("Right clicking will also paste the contents of your clipboard if nothing is selected.")
    while True:
        token = input("Token: ").strip()
        if not token.endswith(".fin"):
            print("What you pasted does not appear to be the token. Copy only what comes after '=>'")
            continue
        new_passport = Passport(token=token)
        if new_passport.test():
            valid_for = new_passport.expires_at - new_passport.current_server_time
            print(f"{Fore.GREEN}Authenticated.{Fore.RESET} Valid for {valid_for} from now")
            PASSPORT = new_passport
            save_passport()
            return True
        print(f"{Fore.RED}Something went wrong, please try again.{Fore.RESET}")


# This gets called a lot, I should probably change how test works some day, but right now it's fine
def is_authenticated() -> bool:
    """Checks if we're authenticated. If a passport file exists and is invalid, it will be deleted"""
    return bool(PASSPORT) and PASSPORT.test()


def seconds_left() -> float:
    """Returns the TTL in seconds"""
    return PASSPORT.seconds_left()


def get_auth_headers():
    """Returns all headers needed to authorize. Does not check validity."""
    if not PASSPORT:
        return None
    return {"Authorization": f"Bearer {PASSPORT.rawlogs_passport}"}



from ss13_tools.auth import interactive


def main():
    """Main"""
    print("Welcome to the token testing service, how may we help you today?")
    interactive()


if __name__ == "__main__":
    main()



from colorama import init as colorama_init

from .key_tools import canonicalize, user_exists
from .__main__ import main  # noqa: F401

colorama_init()

__all__ = [
    'canonicalize',
    'user_exists'
]

del colorama_init



from colorama import init as colorama_init

from .tg import Passport, interactive, is_authenticated, save_passport,\
                  load_passport, create_from_token, get_auth_headers, seconds_left
from .__main__ import main  # noqa: F401


colorama_init()
load_passport()

__all__ = [
    'Passport',
    'interactive',
    'is_authenticated',
    'save_passport',
    'load_passport',
    'create_from_token',
    'seconds_left',
    'get_auth_headers',
]

del colorama_init
